/*
   Class Name:

      CTerrain.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class is a base class used to load and save a terrain.
*/


#include"CTerrain.h"
#include<stdio.h>


CTerrain::CTerrain()
{
   // Initialize member variables.
   m_heightScale = 0.0f;
   m_vertex = NULL;
   m_color = NULL;
   m_texCoords = NULL;
   m_totalVerts = 0;
   m_totalTriangles = 0;
}


CTerrain::~CTerrain()
{
   // Release all resources.
   ShutDown();
}


bool CTerrain::LoadMap(char *filename)
{
   FILE *fp;

   // Error checking.
   if(!filename) return false;

   // If there is any map data release it.
   if(m_heightMap.mapData != NULL) ShutDown();

   // Open the file.
   fp = fopen(filename, "rb");
   if(!fp) return false;

   // Read in the size and the scale values.
   fread(&m_heightMap.size, 1, sizeof(m_heightMap.size), fp);
   fread(&m_heightScale, 1, sizeof(m_heightScale), fp);

   // Initialize the map data variable.
   m_heightMap.mapData = new unsigned char[m_heightMap.size * m_heightMap.size];

   // Error checking.
   if(m_heightMap.mapData == NULL) return false;

   // Read in the entire file into the array.
   fread(m_heightMap.mapData, 1, m_heightMap.size * m_heightMap.size, fp);
   
   CreateTerrainMesh();

   // Close the file.
   fclose(fp);
   return true;
}


void CTerrain::CreateTerrainMesh()
{
   // Index counters.
   int i = 0, c = 0, t = 0;

   if(m_color) delete[] m_color;
   if(m_vertex) delete[] m_vertex;
   if(m_texCoords) delete[] m_texCoords;

   m_totalVerts = 0;
   m_totalTriangles = 0;

   //Sizes we need for the points/color and tex coords.
   int size = ((m_heightMap.size - 1) *
              (m_heightMap.size - 1) * 6) * 3;
   int tSize = ((m_heightMap.size - 1) *
              (m_heightMap.size - 1) * 6) * 2;

   // Create enough space for all vertex points and their colors.
   m_vertex = new float[size];
   m_color = new float[size];
   m_texCoords = new float[tSize];

   // Loop through and generate a grid.  Use the height map
   // when setting the Y to create the terrain mesh.  Create 2 triangles
   // each iteration.
   for(int z = 0; z < m_heightMap.size - 1; z++)
	   {
			for(int x = 0; x < m_heightMap.size - 1; x++)
			   {
			      // Calculate texture coords for these tris.
			      float left = (float)x / m_heightMap.size;
			      float right = ((float)x + 1) / m_heightMap.size;
				   float bottom = (float)z / m_heightMap.size;
				   float top = (float)(z + 1) / m_heightMap.size;

               // V1.
			      m_texCoords[t++] = left;
			      m_texCoords[t++] = bottom;
			      m_color[c++] = GetHeight(x, z);
			      m_color[c++] = GetHeight(x, z);
			      m_color[c++] = GetHeight(x, z);
			      m_vertex[i++] = x;
			      m_vertex[i++] = GetScaledHeight(x, z);
			      m_vertex[i++] = z;
			      m_totalVerts++;

               // V2.
               m_texCoords[t++] = left;
			      m_texCoords[t++] = top;
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
			      m_vertex[i++] = x;
			      m_vertex[i++] = GetScaledHeight(x, z + 1);
			      m_vertex[i++] = z + 1;
			      m_totalVerts++;

               // V3.
               m_texCoords[t++] = right;
			      m_texCoords[t++] = bottom;
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
			      m_vertex[i++] = x + 1;
			      m_vertex[i++] = GetScaledHeight(x + 1, z);
			      m_vertex[i++] = z;
			      m_totalVerts++;


               // V4.
               m_texCoords[t++] = right;
			      m_texCoords[t++] = bottom;
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
               m_color[c++] = GetHeight(x + 1, z);
			      m_vertex[i++] = x + 1;
			      m_vertex[i++] = GetScaledHeight(x + 1, z);
			      m_vertex[i++] = z;
			      m_totalVerts++;

               // V5.
               m_texCoords[t++] = right;
			      m_texCoords[t++] = top;
               m_color[c++] = GetHeight(x + 1, z + 1);
               m_color[c++] = GetHeight(x + 1, z + 1);
               m_color[c++] = GetHeight(x + 1, z + 1);
			      m_vertex[i++] = x + 1;
			      m_vertex[i++] = GetScaledHeight(x + 1, z + 1);
			      m_vertex[i++] = z + 1;
			      m_totalVerts++;

               // V6.
               m_texCoords[t++] = left;
			      m_texCoords[t++] = top;
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
               m_color[c++] = GetHeight(x, z + 1);
			      m_vertex[i++] = x;
			      m_vertex[i++] = GetScaledHeight(x, z + 1);
			      m_vertex[i++] = z + 1;
			      m_totalVerts++;

               m_totalTriangles += 2;
			   }
	   }

   // Convert between 0 and 1.
	for(int k = 0; k < size; k++)
      {
         if(m_color[k] != 0) m_color[k] /= 255;
      }
}


bool CTerrain::SaveMap(char *filename)
{
	FILE *fp;

   // Error Checking...
   if(m_heightMap.mapData == 0) return false;
   if(filename == 0) return false;

	// Open the file to write it out as an text file.
	fp = fopen(filename, "wb");
	
   // Error checking.
   if(fp == 0) return false;

   // Write out the size and scale values.
   fwrite(&m_heightMap.size, 1, sizeof(m_heightMap.size), fp);
   fwrite(&m_heightScale, 1, sizeof(m_heightScale), fp);

	// Write out all of the data in one big write.
	fwrite(m_heightMap.mapData, 1, m_heightMap.size * m_heightMap.size, fp);
	
	// Close the file.
	fclose(fp);
	return true;
}


void CTerrain::SetHeightScale(float val)
{
   // Set the scale value to want is passed in.
   m_heightScale = val;
}


void CTerrain::SetHeight(unsigned char val, int x, int z)
{
   // Set the value to the specified location.
   m_heightMap.mapData[x + m_heightMap.size * z] = val;
}


unsigned char CTerrain::GetHeight(int x, int z)
{
   // Return the unscaled height value of the specified point.
   return m_heightMap.mapData[x + m_heightMap.size * z];
}


float CTerrain::GetScaledHeight(int x, int z)
{
   // Return the scaled height value of the specified point.
   return (m_heightMap.mapData[x + m_heightMap.size * z] * m_heightScale);
}


void CTerrain::ShutDown()
{
   // Release all resources used by the terrain.
   m_heightScale = 0.0f;
   
   if(m_vertex)
      {
         delete[] m_vertex;
         m_vertex = NULL;
      }
   
   if(m_color)
      {
         delete[] m_color;
         m_color = NULL;
      }

   if(m_texCoords)
      {
         delete[] m_texCoords;
         m_texCoords = NULL;
      }

   if(m_heightMap.mapData)
      {
         delete[] m_heightMap.mapData;
         m_heightMap.mapData = NULL;
         m_heightMap.size = 0;
      }
}


// Copyright July 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com